package org.xbib.event.path;

import org.xbib.datastructures.api.TimeValue;
import org.xbib.datastructures.json.tiny.Json;
import org.xbib.event.Event;
import org.xbib.event.common.AbstractEventManagerService;
import org.xbib.event.common.EventManager;
import org.xbib.event.common.EventManagerService;
import org.xbib.settings.Settings;

import java.io.IOException;
import java.io.UncheckedIOException;
import java.io.Writer;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Future;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.stream.Stream;

public class PathEventManagerService extends AbstractEventManagerService implements EventManagerService {

    private static final Logger logger = Logger.getLogger(PathEventManagerService.class.getName());

    private EventManager eventManager;

    private ExecutorService executorService;

    private Map<Future<?>, PathEventService> eventServiceMap;

    public PathEventManagerService() {
        super();
    }

    @Override
    public PathEventManagerService init(EventManager eventManager) {
        this.eventManager = eventManager;
        Settings settings = eventManager.getSettings();
        this.executorService = eventManager.getExecutorService();
        this.eventServiceMap = new LinkedHashMap<>();
        for (Map.Entry<String, Settings> entry : settings.getGroups("event.path").entrySet()) {
            try {
                String name = entry.getKey();
                Settings definition = entry.getValue();
                if (definition.getAsBoolean("enabled", true)) {
                    Path path = Paths.get(definition.get("path", "/var/tmp/" + name));
                    TimeValue lifetime = definition.getAsTime("lifetime", TimeValue.timeValueHours(72));
                    String eventType = definition.get("type", "path");
                    createQueue(name, path);
                    PathEventService pathEventService = new PathEventService(this, name, path, eventType, lifetime);
                    add(pathEventService);
                    pathEventService.drainIncoming();
                } else {
                    logger.log(Level.WARNING, "path service definition not enabled in configuration");
                }
            } catch (Exception e) {
                logger.log(Level.SEVERE, "unable to create event path service " + entry.getKey() + ", reason " + e.getMessage(), e);
            }
        }
        return this;
    }

    public void add(PathEventService pathEventService) {
        Future<?> future = executorService.submit(pathEventService);
        eventServiceMap.put(future, pathEventService);
        logger.log(Level.INFO, "path event service " + pathEventService + " added");
    }

    public void publish(String eventType, Path path) {
        Event event = EventManager.eventOf(eventType, null, null, path);
        eventManager.publish(event);
    }

    public void failEvent(String eventType, Path path) {
        try {
            Event event = EventManager.eventOf(eventType, null, null, path);
            event.fail();
        } catch (IOException e) {
            logger.log(Level.SEVERE, "unable to fail event because of  " + e.getMessage(), e);
        }
    }

    @Override
    public void shutdown() throws IOException {
        logger.log(Level.INFO, "shutting down all path event services");
        eventServiceMap.forEach((k, v) -> {
            k.cancel(true);
            try {
                v.close();
            } catch (IOException e) {
                logger.log(Level.SEVERE, e.getMessage(), e);
            }
        });
    }

    public void destroy() {
        logger.log(Level.INFO, "shutting down and destroying all path event files");
        eventServiceMap.forEach((k, v) -> {
            k.cancel(true);
            try {
                v.close();
                v.destroy();
            } catch (IOException e) {
                logger.log(Level.SEVERE, e.getMessage(), e);
            }
        });
    }

    public boolean putIfNotExists(Path path, String key, Map<String,Object> map) throws IOException {
        if (!exists(path, key, ".json")) {
            return put(path, key, map);
        } else {
            return false;
        }
    }

    public boolean put(Path path, String key, Map<String,Object> map) throws IOException {
        return put(path, key, ".json", Json.toString(map));
    }

    public boolean put(Path path, String key, String suffix, String string) throws IOException {
        String keyFileName = key + suffix;
        if (Files.exists(path.resolve(Event.INCOMING).resolve(keyFileName)) ||
                Files.exists(path.resolve(Event.SUCCESS).resolve(keyFileName))) {
            logger.log(Level.WARNING, "key " + key + " already exists");
            return false;
        }
        Path p = path.resolve(Event.INCOMING).resolve(keyFileName);
        try (Writer writer = Files.newBufferedWriter(p)) {
            writer.write(string);
        }
        // obligatory purge. This is hacky.
        eventServiceMap.forEach((k, v) -> {
            try {
                v.purge();
            } catch (IOException e) {
                throw new UncheckedIOException(e);
            }
        });
        return true;
    }

    public boolean exists(Path path, String key, String suffix) {
        String keyFileName = key + suffix;
        return Files.exists(path.resolve(Event.INCOMING).resolve(keyFileName)) ||
                Files.exists(path.resolve(Event.SUCCESS).resolve(keyFileName));
    }

    public long sizeOfIncoming(Path path) throws IOException {
        return sizeOf(path.resolve(Event.INCOMING));
    }

    public long sizeOfSuccess(Path path) throws IOException {
        return sizeOf(path.resolve(Event.SUCCESS));
    }

    public long sizeOfFail(Path path) throws IOException {
        return sizeOf(path.resolve(Event.FAIL));
    }

    public static long sizeOf(Path path) throws IOException {
        try (Stream<Path> stream = Files.find(path, 1, (p, basicFileAttributes) -> Files.isRegularFile(p))) {
            return stream.count();
        }
    }

    private static void createQueue(String name, Path p) throws IOException {
        logger.log(Level.FINE, "creating queue " + name + " at " + p);
        if (!Files.exists(p)) {
            Files.createDirectories(p);
        }
        for (String s: List.of(Event.INCOMING, Event.SUCCESS, Event.FAIL)) {
            Path dir = p.resolve(s);
            if (!Files.exists(dir)) {
                logger.log(Level.FINE, "creating queue " + name + " dir " + dir);
                Files.createDirectories(dir);
            }
        }
    }
}
